"""Handle yaml configuration files.

Read a yaml file and convert it into a dict.
The format is a bit more restrictive than yaml, you cannot build list of Objects,
only list of values (int, float, bool, str). "." cannot be used inside a key name.

Also some specific keys are used by this library in the root level. See README.md about this.
"""

import copy
import os
import pathlib
import re
import warnings
from typing import Union, cast

import yaml

Value = bool | int | float | str

Config = dict[str, Union[Value, list[Value], "Config"]]


def load_config(config_file: str | os.PathLike) -> Config:
    """Load a configuration from a given file.

    Args:
        config_file (str | os.PathLike): The configuration file name

    Returns:
        Config: The loaded configuration. Parsing is not performed at loading time.
    """
    path = pathlib.Path(config_file)
    cfg: Config = yaml.safe_load(path.read_text(encoding="utf-8"))

    try:
        defaults = cast("str | list[str]", cfg.pop("__default__"))
    except KeyError:
        return cfg

    try:
        new_key_policy = cfg.pop("__new_key_policy__")
    except KeyError:
        new_key_policy = "warn"

    if isinstance(defaults, str):
        defaults = [defaults]

    default_cfg: Config = {}
    for default_path in defaults:
        if pathlib.Path(default_path).is_absolute() or default_path[0] != ".":  # Absolute or from cwd
            default_cfg = merge(default_cfg, load_config(default_path), "pass")
        else:  # Relative to the current config
            default_cfg = merge(default_cfg, load_config(path.parent / default_path), "pass")

    return merge(default_cfg, cfg, new_key_policy)


def save_config(cfg: Config, config_file: str | os.PathLike) -> None:
    """Save the configuration.

    Args:
        cfg (Config): Configuration to save
        config_file (str | os.PathLike): Save to this destination
    """
    path = pathlib.Path(config_file)
    path.write_text(yaml.safe_dump(cfg), encoding="utf-8")


def config_flatten(cfg: Config) -> dict[str, Value | list[Value]]:
    """Flatten a config.

    {
        "hello": {
            "world": True,
            "values": {
                "train": [1, 2, 3],
                "test": 3.5,
            }
        }
    }

    =>

    {
        "hello.world": True,
        "hello.values.train": [1, 2, 3],
        "hello.values.test": 3.5,
    }

    Args:
        cfg (Config): A config to flatten

    Returns:
        Config: The flattened config
    """
    flattened: dict[str, Value | list[Value]] = {}
    _flatten(cfg, flattened)

    return flattened


def _flatten(cfg: Config, flattened: dict[str, Value | list[Value]], prefix: str = "") -> None:
    for key, value in cfg.items():
        true_key = f"{prefix}.{key}" if prefix else key

        # If dict: continue with the new prefix
        if isinstance(value, dict):
            _flatten(value, flattened, true_key)
            continue

        # Else: Let's register the value
        if true_key in flattened:
            raise ValueError(f"Duplicate key: {true_key}. Can't override it")
        flattened[true_key] = value


def config_unflatten(cfg: dict[str, Value | list[Value]]) -> Config:
    """Unflatten a config (Reverse flatten).

    Key containing "." are split. And sub dictionaries are created.

    Args:
        cfg (dict[str, Value | list[Value]]): The config to unflatten.

    Returns:
        Config: The unflattened config
    """
    unflattened_cfg: Config = {}
    for key, value in cfg.items():
        current_cfg = unflattened_cfg

        keys = key.split(".")

        for k in keys[:-1]:
            if k not in current_cfg:
                current_cfg[k] = {}

            next_cfg = current_cfg[k]
            if isinstance(next_cfg, dict):
                current_cfg = next_cfg
            else:
                raise ValueError(f"Duplicated key: {key}. Can't override it. (Found a value for {k})")  # noqa: TRY004

        k = keys[-1]
        current_cfg[k] = value

    return unflattened_cfg


def merge(cfg_1: Config, cfg_2: Config, new_key_policy="warn") -> Config:
    """Merge cfg_2 into cfg_1.

    cfg_1 is not modified. (Nor cfg_2)

    Args:
        cfg_1 (Config): First configuration
        cfg_2 (Config): Second configuration
        new_key_policy (str): Policy when adding new keys
            "raise": Raise KeyError when a key in cfg_2 is not in cfg_1
            "warn": Add the new key but raise a warning
            "pass": Add the new key

    Returns:
        Config: The merged configuration
    """
    if new_key_policy not in ["raise", "warn", "pass"]:
        raise ValueError(f"Unknown `new_key_policy`: {new_key_policy} not in [raise, warn, pass].")

    cfg = copy.deepcopy(cfg_1)

    for key, value in cfg_2.items():
        value = copy.deepcopy(value)  # noqa: PLW2901

        if key not in cfg:
            if key[:2] == "__":  # Allow specific keys to be new
                cfg[key] = value
                continue
            if new_key_policy == "raise":
                raise KeyError(f"Unexpected key when merging configs: {key}")
            if new_key_policy == "warn":
                warnings.warn(f"Adding new key to configuration: {key}", stacklevel=2)
            cfg[key] = value
            continue

        previous_value = cfg[key]

        if not isinstance(value, type(previous_value)):
            warnings.warn(
                f"Key `{key}` has been overloaded with a different type: {type(previous_value)} -> {type(value)}",
                stacklevel=2,
            )
            cfg[key] = value
            continue

        # Same type. If dict, let's merge, if not just replace
        if isinstance(previous_value, dict):
            cfg[key] = merge(previous_value, cast("dict", value), new_key_policy)
        else:
            cfg[key] = value

    return cfg


def convert_if_possible(value: str) -> Value:
    """Try to convert the string to another type if possible.

    Try to convert as much a possible without losing information.
    You should probably use dataclass with dacite to ensure that everything is well typed

    Args:
        value (str): Value to be converted

    Returns:
        Value: Converted value
    """
    try:
        return int(value)
    except ValueError:
        try:
            return float(value)
        except ValueError:
            if value.lower() == "false":
                return False
            if value.lower() == "true":
                return True
            return value


class Parser:
    """Parse a configuration and handles environment variables and self references.

    Attrs:
        NO_PARSE_STRING (str): Add this string to the beginning of the value to prevent parsing
        ENV_VAR_REGEXP, ENV_VAR_REGEXP_BRACKET (str): Regexp for environment variables matching
            A bit too simple, but will do the job
        SELF_REF_REGEXP (str): Regexp for self references. A bit permissive, but fine.
    """

    NO_PARSE_STRING = "!P"
    ENV_VAR_REGEXP = "\\$([a-zA-Z][a-zA-Z1-9_]*)"
    ENV_VAR_REGEXP_BRACKET = "\\$\\{([a-zA-Z][a-zA-Z1-9_]*)\\}"
    SELF_REF_REGEXP = "\\{([a-zA-Z1-9_.]+)\\}"

    def __init__(self, config: Config) -> None:
        self.config = config_flatten(config)
        self.parsing: set[str] = set()
        self.parsed: set[str] = set()

    def parse(self) -> Config:
        """Parse each key/value of the configuration."""
        for key in self.config:  # Flatten, there is only one level
            self.parse_key(key)

        return config_unflatten(self.config)

    def parse_key(self, key: str) -> None:
        """Parse a single key/value of the configuration.

        (can lead to other keys being parsed if the value depends on them)

        Args:
            key (str): Key of the value to parse
        """
        if key in self.parsed:
            return

        if key in self.parsing:
            raise RuntimeError(f"Cyclic references in configuration: Unable to resolve {key}")

        self.parsing.add(key)
        self.config[key] = self.format(self.config[key])

        self.parsing.remove(key)
        self.parsed.add(key)

    def format(self, value: Value | list[Value]) -> Value | list[Value]:
        """Parse and format a value.

        Args:
            value (Value | list[Value]): A value to parse

        Returns:
            Value | list[Value]: Parsed value
        """
        if isinstance(value, list):  # XXX: We do not support list[Config]
            return list(map(self.format, value))  # type: ignore[arg-type]

        if not isinstance(value, str):  # Only strings can be parsed
            return value

        # Allow a simple directive to prevent parsing
        if value[: len(self.NO_PARSE_STRING)] == self.NO_PARSE_STRING:
            return value[len(self.NO_PARSE_STRING) :]

        # Poorly coded
        if "$" in value:
            value = self.replace_env_reference(value)
            if not isinstance(value, str):
                return value

        if "{" in value:
            value = self.replace_self_reference(value)

        return value

    def replace_env_reference(self, value: str) -> Value | list[Value]:
        """Solve environment variables references for the given value.

        Args:
            value (str): String to parse

        Returns:
            Value | list[Value]: Parsed value (converted if possible)
        """

        def env_replace(match):
            env_key = match.group(1)
            if env_key in os.environ:
                return os.environ[env_key]
            warnings.warn(f"Environment variable {env_key} not define", stacklevel=2)
            return ""

        value = re.sub(self.ENV_VAR_REGEXP, env_replace, value)
        value = re.sub(self.ENV_VAR_REGEXP_BRACKET, env_replace, value)
        return convert_if_possible(value)

    def replace_self_reference(self, value: str) -> Value | list[Value]:
        """Solve configuration references for the given value.

        Args:
            value (str): String to parse

        Returns:
            Value | list[Value]: Parsed value (converted if total match)
        """

        def self_replace(match):
            key = match.group(1)
            if key in self.config:
                self.parse_key(key)
                return str(self.config[key])
            warnings.warn(f"Unable to resolve reference {key}.", stacklevel=2)
            return ""

        # Handle full match differently to cast to the right value
        match = re.fullmatch(self.SELF_REF_REGEXP, value)
        if match:
            key = match.group(1)
            if key in self.config:
                self.parse_key(key)
                return copy.deepcopy(self.config[key])  # Keep type if full match

        return re.sub(self.SELF_REF_REGEXP, self_replace, value)



"""Run reproducible experiments from a yaml configuration file.

Expyrun.
=======

Expyrun is a lightweight experiment runner designed to make research workflows
reproducible and structured.

It allows you to:
- Define experiments in YAML configuration files
- Inherit and compose configurations
- Override parameters from the command line
- Resolve environment variables and self-references
- Automatically create structured output directories
- Snapshot the configuration, dependencies, and source code
- Reproduce experiments exactly

Typical workflow
----------------
1. Write a YAML config file containing a special ``__run__`` section.
2. Implement an entry function with signature:

       def main(name: str, config: dict) -> None:

3. Launch your experiment:

       expyrun path/to/config.yml [--debug] [--my.key value]

Expyrun will:
- Build the configuration (including CLI overrides)
- Create a unique experiment directory
- Save the parsed and raw configurations
- Freeze requirements
- Copy the relevant source code (except in debug mode)
- Redirect stdout/stderr to ``outputs.log``
- Execute your entry function inside the experiment directory

This project was originally developed to fit the author's research needs.
It is currently in beta. Contributions and feedback are very welcome.

See the README for full documentation and examples.
"""

import importlib.metadata

__version__ = importlib.metadata.version("expyrun")



"""Run experiment given a configuration file.

The config is expected to have a special `__run__` section.

Command lines arguments are used to overwrite the given configuration.

When an experiment is launched:
1- First the config is built from the config file and args
2- Then the output directory is created.
3- It's then filled with
    a- The built configuration file (with resolved dependencies): config.yml
    b- The built configuration file but unparsed: raw_config.yml
    c- Frozen requirements: frozen_requirements.txt
    d- Copy of all the code in the package of the main function
4- The output directory is registered as the current directory and is set in sys.path.
   You can therefore write directly in the current directory in the main function.
5- Finally the main function is loaded and run.

In DEBUG mode, the code is not copied in the output dir, the code is directly run from the current
directory
"""

import argparse
import atexit
import importlib
import os
import pathlib
import subprocess
import sys
import warnings
from typing import TextIO, cast

from . import config

CODE_FILES_EXTENSIONS = [".py"]


class StdMultiplexer:
    """Patch a writable text stream and multiplexes the outputs to several others.

    Only write and flush are redirected. Other actions are done only on the main stream.
    It will therefore have the same properties as the main stream.
    """

    def __init__(self, main_stream: TextIO, ios: list[TextIO]):
        self.main_stream = main_stream
        self.ios = ios

    def write(self, string: str) -> int:
        """Write to all the streams."""
        ret = self.main_stream.write(string)

        for io_ in self.ios:
            io_.write(string)

        return ret

    def flush(self) -> None:
        """Flush all the streams."""
        self.main_stream.flush()

        for io_ in self.ios:
            io_.flush()

    def __getattr__(self, attr: str):  # noqa: ANN204
        """Redirect all other attributes to the main stream."""
        return getattr(self.main_stream, attr)


class StdFileRedirection:
    """Multiplexes stdout and stderr to a file.

    The code could potentially break other libraries trying to redirects sys.stdout and sys.stderr.
    It has been made compatible with Neptune. Any improvements are welcome.
    """

    def __init__(self, path: str | os.PathLike) -> None:
        self.file = pathlib.Path(path).open("w", encoding="utf-8")  # noqa: SIM115
        self.stdout = StdMultiplexer(sys.stdout, [self.file])
        self.stderr = StdMultiplexer(sys.stderr, [self.file])
        sys.stdout = self.stdout
        sys.stderr = self.stderr
        atexit.register(self.close)

    def close(self):
        """Close the std file redirection (Reset sys.stdout/sys.stderr)."""
        sys.stdout = self.stdout.main_stream
        sys.stderr = self.stderr.main_stream
        self.file.close()


def convert_as(default: config.Value | list[config.Value], arg: str) -> config.Value | list[config.Value]:
    """Convert the argument to the same type found in the config.

    Args:
        default (config.Value | list[config.Value]): Default value in the config file
        arg (str): Argument value

    Returns:
        config.Value | list[config.Value]: The converted value
    """
    if isinstance(default, list):
        values = arg.split(",")
        if len(values) != len(default):
            if len(default) > 0:
                return [cast("config.Value", convert_as(default[0], value)) for value in values]
            return [config.convert_if_possible(value) for value in values]
        return [cast("config.Value", convert_as(default[j], value)) for j, value in enumerate(values)]

    if isinstance(default, bool):
        if arg.lower() in {"false", "0"}:
            return False
        if arg.lower() in {"true", "1"}:
            return True
        raise ValueError(f"Unable to convert to boolean: {arg}")

    converted: config.Value = arg
    if isinstance(default, int):
        converted = int(arg)
    elif isinstance(default, float):
        converted = float(arg)
    elif default is None:
        converted = config.convert_if_possible(arg)

    return converted


def build_config(config_file: str, args: list[str]) -> config.Config:
    """Build the configuration given a config file and extra args.

    Args:
        config_file (str): path to configuration file
        args (list[str]): List of additional args
            Expected format: [--my.entire.key, value, --my.other.key, other_value, ...]
            This feature support only simple types (converted with `convert_as`)

    Returns:
        Config: The config for this run
    """
    cfg = config.load_config(config_file)
    flatten = config.config_flatten(cfg)

    # Parse args to build a cfg
    if len(args) % 2 != 0:
        raise ValueError("# Args should be even => missing a value or a key")

    key = ""
    for i, arg in enumerate(args):
        if i % 2 == 0:
            if arg[:2] != "--":
                raise ValueError("Expected key format: --my.entire.key")
            key = arg[2:]
            continue

        if key not in flatten:
            raise KeyError(f"Unexpected key when merging args: {key}")

        flatten[key] = convert_as(flatten[key], arg)

    return config.config_unflatten(flatten)


def _code_copy(src_path: pathlib.Path, dest_path: pathlib.Path) -> None:
    """Probably inefficient folder search but do the job."""
    dest_path = dest_path / src_path.name
    if dest_path.exists():
        raise RuntimeError("Copying files into an existing location...")

    if src_path.is_file():
        if src_path.suffix in CODE_FILES_EXTENSIONS:
            dest_path.write_bytes(src_path.read_bytes())
        return

    if not src_path.is_dir():
        warnings.warn(f"Unhandled path in code duplication: {src_path}", stacklevel=2)
        return

    dest_path.mkdir()  # Can create empty dir, but fine
    for child_path in src_path.iterdir():
        _code_copy(child_path, dest_path)


def duplicate_code(code_dir: pathlib.Path, output_dir: pathlib.Path, module_name: str) -> None:
    """Duplicate the package containing the running code.

    Args:
        code_dir (Path): Path to the code
        output_dir (Path): Path to where copy the code
        module_name (str): Name of the main module to run
    """
    package_path = code_dir / module_name.split(".", maxsplit=1)[0]
    _code_copy(package_path, output_dir)


def save_requirements(output_dir: pathlib.Path) -> None:
    """Save the requirements using pip freeze into a requirements.txt."""
    # XXX: Not robust, could be improved?
    subprocess.run("pip freeze > requirements.txt", check=False, cwd=output_dir, shell=True)  # noqa: S602, S607
    # OLD: Importing pip blocks setuptools (cf : https://github.com/pypa/setuptools/issues/3044)
    # >> from pip._internal.operations import freeze
    # >> (output_dir / "requirements.txt").write_text("\n".join(freeze.freeze()), encoding="utf-8")


def main(cfg: config.Config, debug: bool) -> None:
    """Prepare and launch the experiment.

    Args:
        cfg (dict): Configuration
        debug (bool): Use DEBUG mode
    """
    # Save the current cwd in case it is really needed (Will be changed in a few lines)
    os.environ["EXPYRUN_CWD"] = str(pathlib.Path.cwd())

    raw_cfg = cfg
    cfg = config.Parser(cfg).parse()  # Resolve self and env references

    # Load the run data
    module_name, func_name = cast("dict[str, str]", cfg["__run__"])["__main__"].split(":")
    experiment_name = cast("dict[str, str]", cfg["__run__"])["__name__"]
    output_dir = pathlib.Path(cast("dict[str, str]", cfg["__run__"])["__output_dir__"])
    code_dir = pathlib.Path(cast("dict[str, str]", cfg["__run__"]).get("__code__", str(pathlib.Path.cwd())))

    # Set output_dir as an absolute path
    output_dir = output_dir.absolute()
    cast("dict[str, str]", raw_cfg["__run__"])["__output_dir__"] = str(output_dir)
    cast("dict[str, str]", cfg["__run__"])["__output_dir__"] = str(output_dir)

    # Compute true output dir
    output_dir = output_dir / "DEBUG" / experiment_name / "exp.0" if debug else output_dir / experiment_name / "exp.0"

    i = 0
    while output_dir.exists():
        i += 1
        output_dir = output_dir.with_suffix(f".{i}")

    # Create the true output dir and fill it
    pathlib.Path.mkdir(output_dir, parents=True, exist_ok=False)

    if debug:  # In debug mode, do not copy the code nor the requirements
        sys.path.insert(0, str(code_dir))
    else:
        save_requirements(output_dir)
        duplicate_code(code_dir, output_dir, module_name)
        cast("dict[str, str]", raw_cfg["__run__"])["__code__"] = str(output_dir)
        cast("dict[str, str]", cfg["__run__"])["__code__"] = str(output_dir)
        sys.path.insert(0, str(output_dir))

    config.save_config(cfg, output_dir / "config.yml")
    config.save_config(raw_cfg, output_dir / "raw_config.yml")

    # Execute inside output_dir
    os.chdir(output_dir)

    # Redirects logs
    StdFileRedirection("outputs.log")

    # Find the main function. Should take the name and the configuration as input
    module = importlib.import_module(module_name)
    _main = getattr(module, func_name)

    # Launch the experiment
    cfg.pop("__run__")
    _main(experiment_name, cfg)


def entry_point() -> None:
    """Expyrun entry point."""
    parser = argparse.ArgumentParser(
        prog="expyrun",
        description=(
            "Run reproducible experiments from a YAML configuration file.\n\n"
            "The configuration file must contain a special '__run__' section "
            "defining the entry function, experiment name, and output directory."
        ),
        epilog=(
            "Examples:\n"
            "  expyrun config.yml\n"
            "  expyrun config.yml --debug\n"
            "  expyrun config.yml --training.lr 0.001\n"
            "  expyrun config.yml --model.size 64 --training.batch_size 32\n\n"
            "When overriding list values, use comma-separated syntax:\n"
            "  --model.layers 64,128,256\n\n"
            "Types are inferred from the configuration file.\n"
            "New keys are not allowed from the command line."
        ),
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        "config",
        help=("Path to the YAML configuration file.\nThe file must define a '__run__' section."),
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help=(
            "Enable DEBUG mode.\n"
            "In this mode, the source code is NOT copied and is executed directly "
            "from the current working directory."
        ),
    )
    parser.add_argument(
        "args",
        nargs=argparse.REMAINDER,
        help=(
            "Additional arguments overriding configuration values.\n"
            "Format: --my.entire.key value\n"
            "Example: --training.lr 0.01\n"
            "         --model.layers 64,128,256"
        ),
    )

    args = parser.parse_args()
    # Hacky counter to the behavior of argparse. Just don't use debug as a config key
    # if --debug is specified after the config file, it will end up with additional args...
    if not args.debug and "--debug" in args.args:
        args.args.remove("--debug")
        args.debug = True

    main(build_config(args.config, args.args), args.debug)
